/* usbcdc.c
* Copyright (C) 2005  Benedikt Sauter
* Copyright (C) 2008  Manuel Stahl
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "usbcdc.h"
#include "../../tiny/usbnapi.h"

#include <avr/pgmspace.h>
#include <stdlib.h>#include "../../../led.h"



#define CS_INTERFACE          0x24
#define CS_ENDPOINT           0x25

enum usb_cdc_function_descriptor {
	FD_HEADER,
	FD_CALL_MGMT,
	FD_ABSTR_CTRL_MGMT,
	FD_DIR_LINE_MGMT,
	FD_TEL_RING,
	FD_TEL_CALL_REP,
	FD_UNION,
//...
	FD_USB_TERMINAL =     0x09,
	FD_NET_CHAN_TERMINAL,
	FD_PROTOCOL_UNIT,
	FD_EXTENSION_UNIT,
	FD_MULTI_CHAN_MGMT,
	FD_CAPI_CTRL_MGMT,
	FD_ETHERNET_NET,
	FD_ATM_NET
};

#ifdef USB_CDC_ACM
	#define USB_CDC_CLASS_DESC_LENGTH 19

	enum usb_cdc_class_request {
		SEND_ENCAPSULATED_COMMAND = 0,
		GET_ENCAPSULATED_RESPONSE,
		SET_COMM_FEATURE,
		GET_COMM_FEATURE,
		CLEAR_COMM_FEATURE,
		SET_LINE_CODING = 0x20,
		GET_LINE_CODING,
		SET_CONTROL_LINE_STATE,
		SEND_BREAK
	};

#endif
#ifdef USB_CDC_ETH
	#define USB_CDC_CLASS_DESC_LENGTH 19
#endif

struct usb_cdc_configuration_descriptor {
	USB_ConfigDescriptor_t		config;
	USB_InterfaceDescriptor_t	interface_ctrl;
	uint8_t	class_desc[USB_CDC_CLASS_DESC_LENGTH];
	USB_EndpointDescriptor_t	ep3_in;
	USB_InterfaceDescriptor_t	interface_data;
	USB_EndpointDescriptor_t	ep1_out;
	USB_EndpointDescriptor_t	ep1_in;
};
typedef struct usb_cdc_configuration_descriptor USB_CDC_ConfigDescriptor_t;


/******************* Device Descriptor ***********************/

USB_DeviceDescriptor_t cdc_device_descriptor PROGMEM = {
		0x12,       // 18 length of device descriptor
		DEVICE,     // descriptor type = device descriptor
    	0x0110,     // version of USB spec. ( e.g. 1.1)
    	CLS_CDC,  	// device class
    	0x00,       // device subclass
    	0x00,       // protocol code
    	0x08,       // depth of ep0 FIFO in byte (e.g. 8)
    	USB_VENDOR_ID,     // vendor id
    	USB_PRODUCT_ID,    // product id
    	0x0100,     // revision id (e.g 1.02)
    	0x01,       // index of manuf. string
    	0x02,       // index of product string
    	0x00,       // index of ser. number
    	0x01        // number of configs
};

/******************* Configuration descriptor *****************/

USB_CDC_ConfigDescriptor_t cdc_config_desc PROGMEM = {
	{ // ConfigurationDescriptor
		sizeof(USB_ConfigDescriptor_t),		// 9 length of this descriptor
		CONFIGURATION,   // descriptor type = configuration descriptor
		sizeof(USB_CDC_ConfigDescriptor_t),	// total length with first interface ...
		0x02,       // number of interfaces
		0x01,       // number if this config. ( arg for setconfig)
		0x00,       // string index for config
		0xA0,       // attrib for this configuration ( bus powerded, remote wakup support)
		0x1A        // power for this configuration in mA (e.g. 50mA)
	}, { // InterfaceDescriptor: CDC CONTROL INTERFACE
		sizeof(USB_InterfaceDescriptor_t),	// 9 length of this descriptor
		INTERFACE,  // descriptor type = interface descriptor
		0x00,       // interface number
		0x00,       // alternate setting for this interface
		0x01,       // number endpoints without 0
		0x02,       // class code
		0x02,       // sub-class code
		0x01,       // protocoll code
		0x00        // string index for interface
	}, { // CDC Class-Specific descriptor
#ifdef USB_CDC_ACM
		0x05,         /* sizeof(usbDescrCDC_HeaderFn): length of descriptor in bytes */
		CS_INTERFACE, /* descriptor type */
		FD_HEADER,    /* header functional descriptor */
		0x10, 0x01,

		4,           /* sizeof(usbDescrCDC_AcmFn): length of descriptor in bytes */
		CS_INTERFACE,/* descriptor type */
		FD_ABSTR_CTRL_MGMT, /* abstract control management functional descriptor */
		0x02,        /* SET_LINE_CODING, GET_LINE_CODING, SET_CONTROL_LINE_STATE    */

		5,           /* sizeof(usbDescrCDC_UnionFn): length of descriptor in bytes */
		CS_INTERFACE,/* descriptor type */
        FD_UNION,    /* union functional descriptor */
		0,           /* CDC_COMM_INTF_ID */
		1,           /* CDC_DATA_INTF_ID */

		5,           /* sizeof(usbDescrCDC_CallMgtFn): length of descriptor in bytes */
		CS_INTERFACE,/* descriptor type */
		FD_CALL_MGMT,/* call management functional descriptor */
		3,           /* allow management on data interface, handles call management by itself */
		1            /* CDC_DATA_INTF_ID */
#endif
	}, { // EndpointDescriptor
		sizeof(USB_EndpointDescriptor_t),	/* sizeof(usbDescrEndpoint) */
		ENDPOINT,    /* descriptor type = endpoint */
		0x83,        /* IN endpoint number 3 */
		0x03,        /* attrib: Interrupt endpoint */
		0x0008,      /* maximum packet size */
		100          /* in ms */
	}, { //Interface Descriptor: CDC DATA INTERFACE
		sizeof(USB_InterfaceDescriptor_t),	/* sizeof(usbDescrInterface): length of descriptor in bytes */
		INTERFACE,   /* descriptor type */
		1,           /* index of this interface */
		0,           /* alternate setting for this interface */
		2,           /* endpoints excl 0: number of endpoint descriptors to follow */
		0x0A,        /* Data Interface Class Codes */
		0,
		0,           /* Data Interface Class Protocol Codes */
		0            /* string index for interface */
	}, { // EndpointDescriptor
		sizeof(USB_EndpointDescriptor_t),	/* sizeof(usbDescrEndpoint) */
		ENDPOINT,    /* descriptor type = endpoint */
		0x01,        /* OUT endpoint number 1 */
		BULK,        /* attrib: Bulk endpoint */
		0x0008,      /* maximum packet size 8->6 */
		0            /* in ms */
	}, { // EndpointDescriptor
		sizeof(USB_EndpointDescriptor_t),	/* sizeof(usbDescrEndpoint) */
		ENDPOINT,    /* descriptor type = endpoint */
		0x81,        /* IN endpoint number 1 */
		BULK,        /* attrib: Bulk endpoint */
		0x0008,      /* maximum packet size */
		0            /* in ms */
	}
};

USB_CDC_LineCoding_t lineCoding = {0x00002580, 0, 0, 8};
uint8_t EP1out_buf[8];
//                data, cmd,  ctlr,
EPInfo EP3in  = { TXD1, TXC1, EPC1, 8, 0, EP_MEM, 0, 0, 0 };
EPInfo EP1out = { RXD1, RXC1, EPC2, 8, 0, EP_MEM, 0, 0, EP1out_buf };
EPInfo EP1in  = { TXD2, TXC2, EPC3, 8, 0, EP_MEM, 0, 0, 0 };


void USB_CDC_EP1_receive(uint8_t count)
{
	EP1out.bufSize = count;
	USBNReceive(&EP1out);
	USB_CDC_rxCallback(EP1out.buf, count);
}

void USB_CDC_tx(uint8_t* data, uint8_t len)
{
	EP1in.buf = data;
	EP1in.bufSize = len;
	EP1in.bufIndex = 0;
	EP1in.bufType = EP_MEM;
	USBNTransmit(&EP1in);
}

void USB_CDC_init(void)
{
	USBNInit((uint8_t*)&cdc_device_descriptor,(uint8_t*)&cdc_config_desc);
}

/*************** USB CDC class requests  **************/

#ifdef USB_CDC_ACM
void USBNDecodeClassRequest(DeviceRequest_t *req)
{
	static prog_uint8_t serialNotification[10] = {0xa1,0x20,0,0,0,0,2,0,3,0};

//	uint8_t stat = USBNRead(RXS0);
	EP3in.bufType = EP_MEM;
	switch(req->bRequest) {
		case SET_LINE_CODING:
			EP0rx.buf = (uint8_t*)&lineCoding;
			EP0rx.bufSize = sizeof(USB_CDC_LineCoding_t);
			USBNReceive(&EP0rx);

			USB_CDC_setLineCoding(&lineCoding);
//			lineCoding.dwDTERrate = 0x2580;		// override baudrate with 9600

			USBNWrite(TXC0,FLUSH);
			USBNBurstWrite(TX_TOGL+TX_EN);
		break;
		case GET_LINE_CODING:
			EP0tx.buf = (uint8_t*)&lineCoding;
			EP0tx.bufSize = sizeof(USB_CDC_LineCoding_t);
			EP0tx.bufIndex = 0;
			EP0tx.bufType = EP_MEM;
			USBNTransmit(&EP0tx);
		break;
		case SET_CONTROL_LINE_STATE:
			//TODO: set line state according to req->wValue

			EP3in.bufIndex = 0;
			EP3in.bufSize = 8;
			EP3in.bufType = EP_FLASH;
			EP3in.buf = serialNotification;
			USBNTransmit(&EP3in);

			USBNWrite(TXC0,FLUSH);
			USBNBurstWrite(TX_TOGL+TX_EN);
		break;
//		default:
//			LED_on();
		break;
	}
	USBNWrite(RXC0,FLUSH);
}

/*************** USB CDC configuration   **************/

void USBNSetConfiguration(DeviceRequest_t *req) {
	// define EPaddresses
	// TODO: save the configuration index

	USBNWrite(TXC1, FLUSH);
	USBNWrite(EPC1, EP_EN + 0x03);	// FIFO1: EP3 interrupt in

	USBNWrite(RXC1, FLUSH);
	USBNWrite(EPC2, EP_EN + 0x01);	// FIFO2: EP1 bulk out
	USBNWrite(RXC1, RX_EN);
	USBNCallbackFIFORX1(&USB_CDC_EP1_receive);

	USBNWrite(TXC2, FLUSH);
	USBNWrite(EPC3, EP_EN + 0x01);	// FIFO3: EP1 bulk in

	USBNWrite(TXC0, TX_TOGL + TX_EN); //enable the TX (DATA1)
}

#endif
